@@ -6,6 +6,8 @@ import warnings
 import matplotlib.pyplot as plt
 import astropy as ap
 import astropy.units as u
+import mpmath as mp
+mp.dps=15; mp.pretty=True
 
 g=0.004317#newton's G in units of km/s, pc, Msun
 g_dim=g*u.km**2*u.pc/u.s**2/u.M_sun#now as an astropy.units quantity
@@ -26,7 +28,12 @@ def get_nfw_scale(triangle,h,m_triangle,c_triangle):#r_triangle, scale radius r_
     else:
         r_triangle=(2.*g*m_triangle/triangle*(1.e+6**2)/(h*100.)**2)**(1/3)#r_triangle in units of pc, where triangle is overdensity factor = [M_triangle / (4*pi*r_triangle**3)] / rho_crit_0, where rho_crit_0 = 3H_0^2/(8*pi*G), H_0 is hubble constant, m_triangle is given in units of Msun
     r_scale=r_triangle/c_triangle#scale radius in same units as r_triangle, where concentration is defined as c_triangle=r_triangle/r_scale
-    return r_triangle,r_scale,gc*m_triangle/4./np.pi/r_scale**3
+    rho_scale=gc*m_triangle/4./np.pi/r_scale**3
+    if type(m_triangle) is ap.units.quantity.Quantity:    
+        phi0=(-4.*np.pi*g_dim*rho_scale*r_scale**2).to(u.km*u.km/u.s/u.s)
+    else:
+        phi0=-4.*np.pi*g*rho_scale*r_scale**2
+    return r_triangle,r_scale,rho_scale,phi0
 
 def get_dehnen_core_scale(triangle,h,m_triangle,c_triangle):#r_triangle, scale radius r_s and scale density rho_s, of NFW halo, units of pc and U(m_triangle) / u(r_scale)**3
     gc=get_dehnen_core_gc(c_triangle)
@@ -35,7 +42,12 @@ def get_dehnen_core_scale(triangle,h,m_triangle,c_triangle):#r_triangle, scale r
     else:
         r_triangle=(2.*g*m_triangle/triangle*(1.e+6**2)/(h*100.)**2)**(1/3)#r_triangle in units of pc, where triangle is overdensity factor = [M_triangle / (4*pi*r_triangle**3)] / rho_crit_0, where rho_crit_0 = 3H_0^2/(8*pi*G), H_0 is hubble constant, m_triangle is given in units of Msun
     r_scale=r_triangle/c_triangle#scale radius in same units as r_triangle, where concentration is defined as c_triangle=r_triangle/r_scale
-    return r_triangle,r_scale,gc*m_triangle/(4./3.)/np.pi/r_scale**3
+    rho_scale=gc*m_triangle/(4./3.)/np.pi/r_scale**3
+    if type(m_triangle) is ap.units.quantity.Quantity:    
+        phi0=(-2./3.*np.pi*g_dim*rho_scale*r_scale**2).to(u.km*u.km/u.s/u.s)
+    else:
+        phi0=-2./3.*np.pi*g*rho_scale*r_scale**2
+    return r_triangle,r_scale,rho_scale,phi0
 
 def get_dehnen_cusp_scale(triangle,h,m_triangle,c_triangle):#r_triangle, scale radius r_s and scale density rho_s, of NFW halo, units of pc and U(m_triangle) / u(r_scale)**3
     gc=get_dehnen_cusp_gc(c_triangle)
@@ -44,7 +56,12 @@ def get_dehnen_cusp_scale(triangle,h,m_triangle,c_triangle):#r_triangle, scale r
     else:
         r_triangle=(2.*g*m_triangle/triangle*(1.e+6**2)/(h*100.)**2)**(1/3)#r_triangle in units of pc, where triangle is overdensity factor = [M_triangle / (4*pi*r_triangle**3)] / rho_crit_0, where rho_crit_0 = 3H_0^2/(8*pi*G), H_0 is hubble constant, m_triangle is given in units of Msun
     r_scale=r_triangle/c_triangle#scale radius in same units as r_triangle, where concentration is defined as c_triangle=r_triangle/r_scale
-    return r_triangle,r_scale,gc*m_triangle/(4./2.)/np.pi/r_scale**3
+    rho_scale=gc*m_triangle/(4./2.)/np.pi/r_scale**3
+    if type(m_triangle) is ap.units.quantity.Quantity:    
+        phi0=(-2.*np.pi*g_dim*rho_scale*r_scale**2).to(u.km*u.km/u.s/u.s)
+    else:
+        phi0=-2.*np.pi*g*rho_scale*r_scale**2
+    return r_triangle,r_scale,rho_scale,phi0
 
 def get_abg_triangle_scale(triangle,h,m_triangle,c_triangle,alpha,beta,gamma):#r_triangle, scale radius r_s and scale density, rho_s, of abg halo, given triangle parameters, units of U(m_triangle)/U(r_triangle)**3
     if type(m_triangle) is ap.units.quantity.Quantity:
@@ -52,13 +69,14 @@ def get_abg_triangle_scale(triangle,h,m_triangle,c_triangle,alpha,beta,gamma):#r
     else:
         r_triangle=(2.*g*m_triangle/triangle*(1.e+6**2)/(h*100.)**2)**(1/3)#r_triangle in units of pc, where triangle is overdensity factor = [M_triangle / (4*pi*r_triangle**3)] / rho_crit_0, where rho_crit_0 = 3H_0^2/(8*pi*G), H_0 is hubble constant, m_triangle is given in units of Msun
     r_scale=r_triangle/c_triangle#scale radius in same units as r_triangle, where concentration is defined as c_triangle=r_triangle/r_scale
-        
     a=(3.-gamma)/alpha
     b=(beta-gamma)/alpha
     c=(3.-gamma+alpha)/alpha
     z1=-c_triangle**alpha
-    hf1=scipy.special.hyp2f1(a,b,c,z1)  
-    return r_triangle,r_scale,m_triangle*(3.-gamma)/4./np.pi/(r_scale**3)/c_triangle**(3.-gamma)/hf1
+    hf1=scipy.special.hyp2f1(a,b,c,z1)      
+    rho_scale=m_triangle*(3.-gamma)/4./np.pi/(r_scale**3)/c_triangle**(3.-gamma)/hf1
+    phi0=np.nan#not yet implemented
+    return r_triangle,r_scale,rho_scale,phi0
 
 def nfw_mass_density(x,c_triangle):# returns rho_NFW(x) / rho_scale, where x = r / r_triangle
     cx=c_triangle*x #r / r_scale
@@ -77,16 +95,31 @@ def nfw_enclosed_mass(x,c_triangle):# returns enclosed mass M(x) / m_triangle, w
     cx=c_triangle*x #r / r_scale
     return gc*(np.log(1.+cx)-cx/(1.+cx))
 
+def nfw_potential(x,c_triangle): # returns gravitational potential phi(x) / phi(0), where x=r/r_triangle
+    gc=get_nfw_gc(c_triangle)
+    cx=c_triangle*x #r / r_scale
+    return np.log(1.+cx)/cx
+    
 def dehnen_core_enclosed_mass(x,c_triangle):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
     gc=get_dehnen_core_gc(c_triangle)
     cx=c_triangle*x #r / r_scale
     return gc*(cx**3)/(1.+cx)**3
 
+def dehnen_core_potential(x,c_triangle): # returns gravitational potential phi(x) / phi(0), where x=r/r_triangle
+    gc=get_dehnen_core_gc(c_triangle)
+    cx=c_triangle*x #r / r_scale
+    return (1.+3.*cx+2.*cx**2)/(1.+cx)**3
+
 def dehnen_cusp_enclosed_mass(x,c_triangle):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
     gc=get_dehnen_cusp_gc(c_triangle)
     cx=c_triangle*x #r / r_scale
     return gc*(cx**2)/(1.+cx)**2
 
+def dehnen_cusp_potential(x,c_triangle):# returns gravitational potential phi(x) / phi(0), where x = r/r_triangle
+    gc=get_dehnen_cusp_gc(c_triangle)
+    cx=c_triangle*x #r / r_scale
+    return 1./(1.+cx)
+
 def fncore(x,r_core,n_core):#returns f^n(x) for coreNFW model (Read, Walker, Pascal 2018), where x=r/r_triangle, r_core=r_core/r_triangle
     return (np.tanh(np.float64(x)/r_core))**n_core
 
@@ -95,9 +128,16 @@ def cnfw_mass_density(x,c_triangle,r_core,n_core):#returns rho_coreNFW(x) / rho_
     two=2.
     return fncore(x,r_core,n_core)*nfw_mass_density(x,c_triangle)+n_core*fncore(x,r_core,ncorem1)*(1.-fncore(x,r_core,two))/(x**2)/get_nfw_gc(c_triangle)*nfw_enclosed_mass(x,c_triangle)/r_core/(c_triangle**3)
 
-def cnfw_enclosed_mass(x,c_triangle,r_core,n_core):# returns M_cNFW(x) / m_triangle, where x=r/r_triangle, x=r/r_triangle, r_core=(core radius)/ r_triangle
+def cnfw_enclosed_mass(x,c_triangle,r_core,n_core):# returns M_cNFW(x) / m_triangle, where x=r/r_triangle, r_core=(core radius)/ r_triangle
     return fncore(x,r_core,n_core)*nfw_enclosed_mass(x,c_triangle)
 
+def cnfw_potential(x,c_triangle,r_core,n_core):# returns phi_cNFW(x) / phi_cNFW(0), where x=r/r_triangle, r_core=(core radius)/ r_triangle
+    ncorem1=n_core-1.
+    two=2.
+    cx=c_triangle*x #r / r_scale
+    polylog_vec=np.frompyfunc(mp.polylog,2,1)
+    return (fncore(x,r_core,n_core)/cx*(np.log(1.+cx)-cx/(1.+cx)+1./(1.+cx)-1./(1.+c_triangle))+1./c_triangle/r_core*n_core*fncore(x,r_core,ncorem1)*(1.-fncore(x,r_core,two))*(np.log(1.+cx)-np.log(1.+c_triangle)+polylog_vec(2,-cx)-float(mp.polylog(2,-c_triangle))))/(fncore(x,r_core,n_core)*(2.-1./(1.+c_triangle))+n_core*1./c_triangle/r_core*fncore(x,r_core,ncorem1)*fncore(x,r_core,two)*(-np.log(1.+c_triangle)-float(mp.polylog(2,-c_triangle))))
+
 def cnfwt_mass_density(x,c_triangle,r_core,n_core,r_tide,delta):#returns rho_coreNFWtides(x) / rho_0, where x = r/r_triangle
     if ((type(x) is float)|(type(x) is np.float64)):
         if x<r_tide:
@@ -122,6 +162,12 @@ def cnfwt_enclosed_mass(x,c_triangle,r_core,n_core,r_tide,delta):#returns M_cNFW
         val[x>=r_tide]=cnfw_enclosed_mass(r_tide,c_triangle,r_core,n_core)+cnfw_mass_density(r_tide,c_triangle,r_core,n_core)*get_nfw_gc(c_triangle)/(3.-delta)*((c_triangle*r_tide)**3)*(((x[x>=r_tide]/r_tide)**(3.-delta))-1.)
         return val
 
+def cnfwt_potential(x,c_triangle,r_core,n_core):# returns phi_cNFW(x) / phi_cNFW(0), where x=r/r_triangle, r_core=(core radius)/ r_triangle
+    ncorem1=n_core-1.
+    two=2.
+    cx=c_triangle*x #r / r_scale
+    return np.nan
+
 def abg_triangle_mass_density(x,c_triangle,alpha,beta,gamma):# returns rho_abg(x) / rho_scale, where x = r / r_triangle
     cx=c_triangle*x #r / r_scale
     return 1./(cx**gamma)/(1.+cx**alpha)**((beta-gamma)/alpha)
@@ -140,6 +186,9 @@ def abg_triangle_enclosed_mass(x,c_triangle,alpha,beta,gamma):# returns enclosed
     hf2=scipy.special.hyp2f1(a,b,c,z2)
     return ((cx/c_triangle)**(3.-gamma))*hf1/hf2
 
+def abg_triangle_potential(x,c_triangle,alpha,beta,gamma):# returns enclosed mass M_abg(x) / m_triangle, where x = r/r_triangle
+    return np.nan#not yet implemented
+
 def get_plum_scale(luminosity_tot,r_scale):#nu_scale, normalization factor for number density profile
     nu_scale=3.*luminosity_tot/4./np.pi/r_scale**3
     sigma0=luminosity_tot/np.pi/r_scale**2
@@ -184,15 +233,9 @@ def plum_luminosity_density_2d(x):#Sigma(X) / Sigma0, X=R/r_scale
 
 def exp_luminosity_density(x):#nu(x) / nu_scale, x=r/r_scale
     if type(x) is ap.units.quantity.Quantity:#have to work around problems with scipy.special.modstruve working with quantities
-        if x.value<1.e-10:
-            return np.inf
-        else:
-            return scipy.special.kn(0,x.value)
+        return scipy.special.kn(0,x.value)
     else:
-        if x<1.e-10:
-            return np.inf
-        else:
-            return scipy.special.kn(0,x)
+        return scipy.special.kn(0,x)
 
 def exp_luminosity_density_2d(x):#Sigma(X) / Sigma0, X=R/r_scale
     return np.exp(-x)
@@ -234,10 +277,16 @@ def exp_enclosed_luminosity(x):#L(x) / luminosity_tot, x=r/r_scale
         result=1./(3.*np.pi)*x*(3.*np.pi*scipy.special.kn(2,x.value)*scipy.special.modstruve(1,x.value)+scipy.special.kn(1,x.value)*(3.*np.pi*scipy.special.modstruve(2,x.value)-4.*x.value))
         if ((type(x.value) is list)|(type(x.value) is np.ndarray)):
             result[x.value>100]=1.
+        else:
+            if x.value>100:
+                result=1.
     else:
         result=1./(3.*np.pi)*x*(3.*np.pi*scipy.special.kn(2,x)*scipy.special.modstruve(1,x)+scipy.special.kn(1,x)*(3.*np.pi*scipy.special.modstruve(2,x)-4.*x))
         if ((type(x) is list)|(type(x) is np.ndarray)):
             result[x>100]=1.
+        else:
+            if x>100.:
+                result=1.
     return result
 
 def a2bg_enclosed_luminosity(x,beta,gamma):#L(x)/luminosity_tot, x=r/r_scale
@@ -319,7 +368,7 @@ def get_dmhalo(model,**params):
     
     class dmhalo:
         
-        def __init__(self,model=None,triangle=None,h=None,m_triangle=None,c_triangle=None,r_triangle=None,r_core=None,n_core=None,r_tide=None,delta=None,alpha=None,beta=None,gamma=None,rho_scale=None,r_scale=None,v_max=None,r_max=None,mass_density=None,enclosed_mass=None,vcirc=None):
+        def __init__(self,model=None,triangle=None,h=None,m_triangle=None,c_triangle=None,r_triangle=None,r_core=None,n_core=None,r_tide=None,delta=None,alpha=None,beta=None,gamma=None,rho_scale=None,r_scale=None,v_max=None,r_max=None,mass_density=None,enclosed_mass=None,vcirc=None,potential=None,phi0=None):
 
             self.model=model
             self.triangle=triangle
@@ -341,60 +390,82 @@ def get_dmhalo(model,**params):
             self.mass_density=mass_density
             self.enclosed_mass=enclosed_mass
             self.vcirc=vcirc
-
+            self.potential=potential
+            self.phi0=phi0
+            
     if model=='nfw':
         
-        r_triangle,r_scale,rho_scale=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        r_triangle,r_scale,rho_scale,phi0=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
         
         def mass_density(x):
-            return nfw_mass_density(x,params['c_triangle'])
+            return nfw_mass_density(x,params['c_triangle'])#returns mass density rho(x) / rho_scale, where x = r/r_triangle
         def enclosed_mass(x):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
             return nfw_enclosed_mass(x,params['c_triangle'])
-
+        def potential(x):
+            return nfw_potential(x,params['c_triangle'])#returns potential Phi(x)/Phi(0), where x = r/r_triangle, Phi(0) is central potential
+        
     if model=='dehnen_core':
         
-        r_triangle,r_scale,rho_scale=get_dehnen_core_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        r_triangle,r_scale,rho_scale,phi0=get_dehnen_core_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
         
         def mass_density(x):
             return dehnen_core_mass_density(x,params['c_triangle'])
         def enclosed_mass(x):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
             return dehnen_core_enclosed_mass(x,params['c_triangle'])
+        def potential(x):
+            return dehnen_core_potential(x,params['c_triangle'])
 
     if model=='dehnen_cusp':
         
-        r_triangle,r_scale,rho_scale=get_dehnen_cusp_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        r_triangle,r_scale,rho_scale,phi0=get_dehnen_cusp_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
         
         def mass_density(x):
             return dehnen_cusp_mass_density(x,params['c_triangle'])
         def enclosed_mass(x):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
             return dehnen_cusp_enclosed_mass(x,params['c_triangle'])
+        def potential(x):
+            return dehnen_cusp_potential(x,params['c_triangle'])
 
     elif model=='abg':
 
-        r_triangle,r_scale,rho_scale=get_abg_triangle_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'],params['alpha'],params['beta'],params['gamma'])
+        r_triangle,r_scale,rho_scale,phi0=get_abg_triangle_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'],params['alpha'],params['beta'],params['gamma'])
 
         def mass_density(x):
             return abg_triangle_mass_density(x,params['c_triangle'],params['alpha'],params['beta'],params['gamma'])
         def enclosed_mass(x):
             return abg_triangle_enclosed_mass(x,params['c_triangle'],params['alpha'],params['beta'],params['gamma'])
+        def potential(x):
+            return abg_triangle_potential(x,params['c_triangle'])
 
     elif model=='cnfw':#params['r_core'] is core radius
         
-        r_triangle,r_scale,rho_scale=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        r_triangle,r_scale,rho_scale,phi0=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        
+        ncorem1=params['n_core']-1.
+        two=2.
+        if type(params['m_triangle']) is ap.units.quantity.Quantity:    
+            phi0=(-4.*np.pi*g_dim*rho_scale*r_scale**2*(fncore(0.,params['r_core'],params['n_core'])*(2.-1./(1.+params['c_triangle']))+params['n_core']*1./params['c_triangle']/params['r_core']*fncore(0.,params['r_core'],ncorem1)*(1.-fncore(0.,params['r_core'],two))*(-np.log(1.+params['c_triangle'])-float(mp.polylog(2.,-params['c_triangle']))))).to(u.km*u.km/u.s/u.s)
+        else:
+            phi0=-4.*np.pi*g*rho_scale*r_scale**2*(fncore(0.,params['r_core'],params['n_core'])*(2.-1./(1.+params['c_triangle']))+params['n_core']*1./params['c_triangle']/params['r_core']*fncore(0.,params['r_core'],ncorem1)*(1.-fncore(0.,params['r_core'],two))*(-np.log(1.+params['c_triangle'])-float(mp.polylog(2.,-params['c_triangle']))))
         
         def mass_density(x):
             return cnfw_mass_density(x,params['c_triangle'],params['r_core'],params['n_core'])
         def enclosed_mass(x):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
             return cnfw_enclosed_mass(x,params['c_triangle'],params['r_core'],params['n_core'])
+        def potential(x):
+            return cnfw_potential(x,params['c_triangle'],params['r_core'],params['n_core'])
 
     elif model=='cnfwt':#params['r_core'] is core radius / r_triangle, params['r_tide'] is tidal radius / r_triangle
         
-        r_triangle,r_scale,rho_scale=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        r_triangle,r_scale,rho_scale,phi0=get_nfw_scale(params['triangle'],params['h'],params['m_triangle'],params['c_triangle'])
+        phi0=np.nan#not yet implemented
         
         def mass_density(x):
             return cnfwt_mass_density(x,params['c_triangle'],params['r_core'],params['n_core'],params['r_tide'],params['delta'])
         def enclosed_mass(x):# returns enclosed mass M(x) / m_triangle, where x = r/r_triangle
             return cnfwt_enclosed_mass(x,params['c_triangle'],params['r_core'],params['n_core'],params['r_tide'],params['delta'])
+        def potential(x):
+            return cnfwt_potential(x,params['c_triangle'],params['r_core'],params['n_core'],params['r_tide'],params['delta'])
         
     def vcirc(x):# returns circular velocity, km/s
         if type(params['m_triangle']) is ap.units.quantity.Quantity:
@@ -412,17 +483,17 @@ def get_dmhalo(model,**params):
     v_max=vcirc(res.x[0])
 
     if model=='nfw':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     if model=='dehnen_core':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     if model=='dehnen_cusp':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     elif model=='abg':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,alpha=params['alpha'],beta=params['beta'],gamma=params['gamma'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,alpha=params['alpha'],beta=params['beta'],gamma=params['gamma'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     elif model=='cnfw':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,r_core=params['r_core'],n_core=params['n_core'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,r_core=params['r_core'],n_core=params['n_core'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     elif model=='cnfwt':
-        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,r_core=params['r_core'],n_core=params['n_core'],r_tide=params['r_tide'],delta=params['delta'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc)
+        return dmhalo(model=model,triangle=params['triangle'],h=params['h'],m_triangle=params['m_triangle'],c_triangle=params['c_triangle'],r_triangle=r_triangle,r_core=params['r_core'],n_core=params['n_core'],r_tide=params['r_tide'],delta=params['delta'],rho_scale=rho_scale,r_scale=r_scale,v_max=v_max,r_max=r_max,mass_density=mass_density,enclosed_mass=enclosed_mass,vcirc=vcirc,potential=potential,phi0=phi0)
     else:
         raise TypeError('DM halo not properly specified!')
     
@@ -583,7 +654,7 @@ def get_rhalf(model,r_scale,**params):
     else:
         raise ValueError('error in model specification')
     return rhalf_2d,rhalf_3d,nu_scale,ntot
-    
+
 def integrate(bigx,dmhalo,tracer,anisotropy,**params):
     
     if 'component' not in params:
@@ -638,7 +709,7 @@ def integrate(bigx,dmhalo,tracer,anisotropy,**params):
     
     bigsigmasigmalos2,bigsigmasigmarad2,bigsigmasigmatan2,nusigmarad2,nusigmatan2,sigma_proj_los,sigma_proj_rad,sigma_proj_tan,sigma_rad,sigma_tan=np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan
     
-    if min0==max0:
+    if min0>=max0:
         bigsigmasigmalos2,bigsigmasigmarad2,bigsigmasigmatan2,nusigmarad2,nusigmatan2,sigma_proj_los,sigma_proj_rad,sigma_proj_tan,sigma_rad,sigma_tan=0.,0.,0.,0.,0.,0.,0.,0.,0.,0.
         
     else:
@@ -676,7 +747,12 @@ def integrate(bigx,dmhalo,tracer,anisotropy,**params):
         sigma_tan=np.sqrt(nusigmatan2/(tracer.luminosity_density(bigx_tracer)))
     return jeans_integral(sigma_proj_los=sigma_proj_los,sigma_proj_rad=sigma_proj_rad,sigma_proj_tan=sigma_proj_tan,sigma_rad=sigma_rad,sigma_tan=sigma_tan)
 
+
+
+
+
 def integrate_isotropic(bigx,dmhalo,tracer,**params):
+    
     if not 'upper_limit' in params:#default upper limit is infinity, common alternative is dmhalo.r_triangle
         params['upper_limit']=np.inf
     if not 'epsrel' in params:
@@ -685,6 +761,10 @@ def integrate_isotropic(bigx,dmhalo,tracer,**params):
         params['epsabs']=1.49e-8
     if not 'limit' in params:
         params['limit']=50
+
+    class jeans_integral_isotropic:
+        def __init__(self,sigma_proj_los=None):
+            self.sigma_proj_los=sigma_proj_los
         
     def integrand1(x_halo,dmhalo,tracer):
         x_tracer=x_halo*dmhalo.r_triangle/tracer.r_scale# r / r_scale
@@ -693,10 +773,28 @@ def integrate_isotropic(bigx,dmhalo,tracer,**params):
     
     min0=bigx
     max0=params['upper_limit']
-    if type(m_triangle) is ap.units.quantity.Quantity:
-        return 2.*g_dim*dmhalo.m_triangle*scipy.integrate.quad(integrand1,min0,max0,args=(dmhalo,tracer),epsrel=params['epsrel'],epsabs=params['epsabs'])[0]#sigma^2_LOS(X) * Sigma(X) / nu_scale
+    bigx_tracer=bigx*dmhalo.r_triangle/tracer.r_scale
+    
+    bigsigmasigmalos2,bigsigmasigmarad2,bigsigmasigmatan2,nusigmarad2,nusigmatan2,sigma_proj_los,sigma_proj_rad,sigma_proj_tan,sigma_rad,sigma_tan=np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan
+    
+    if min0>=max0:
+        bigsigmasigmalos2,bigsigmasigmarad2,bigsigmasigmatan2,nusigmarad2,nusigmatan2,sigma_proj_los,sigma_proj_rad,sigma_proj_tan,sigma_rad,sigma_tan=0.,0.,0.,0.,0.,0.,0.,0.,0.,0.
+        
     else:
-        return 2.*g*dmhalo.m_triangle*scipy.integrate.quad(integrand1,min0,max0,args=(dmhalo,tracer),epsrel=params['epsrel'],epsabs=params['epsabs'])[0]#sigma^2_LOS(X) * Sigma(X) / nu_scale
+        
+        if type(dmhalo.m_triangle) is ap.units.quantity.Quantity:
+            bigsigmasigmalos2=2.*g_dim*dmhalo.m_triangle*scipy.integrate.quad(integrand1,min0,max0,args=(dmhalo,tracer),epsrel=params['epsrel'],epsabs=params['epsabs'])[0]#sigma^2_los(X) * Sigma(X) / nu_scale
+        else:
+            bigsigmasigmalos2=2.*g*dmhalo.m_triangle*scipy.integrate.quad(integrand1,min0,max0,args=(dmhalo,tracer),epsrel=params['epsrel'],epsabs=params['epsabs'])[0]#sigma^2_los(X) * Sigma(X) / nu_scale
+                
+        sigma_proj_los=np.sqrt(bigsigmasigmalos2/(tracer.luminosity_density_2d(bigx_tracer)*tracer.sigma0/tracer.nu_scale))
+        
+    return jeans_integral_isotropic(sigma_proj_los=sigma_proj_los)
+
+
+
+
+
 
 def projected_virial(x_halo,dmhalo,tracer):#computes integral for Wlos from Errani etal (2018)
     x_tracer=x_halo*dmhalo.r_triangle/tracer.r_scale
@@ -719,7 +817,8 @@ def get_virial(dmhalo,tracer,**params):
     if type(dmhalo.m_triangle) is ap.units.quantity.Quantity:
         vvar=val1[0]*4.*np.pi*g_dim/3.*dmhalo.m_triangle*(dmhalo.r_triangle**2)/tracer.ltotnorm/tracer.r_scale**3
         mu=g_dim*(dmhalo.enclosed_mass(params['lamb']*tracer.rhalf_2d/dmhalo.r_triangle)+tracer.enclosed_luminosity(params['lamb']*tracer.rhalf_2d/tracer.r_scale)*tracer.luminosity_tot*tracer.upsilon/dmhalo.m_triangle)*dmhalo.m_triangle/params['lamb']/tracer.rhalf_2d/vvar
+        return np.sqrt(vvar),mu.value
     else:
         vvar=val1[0]*4.*np.pi*g/3.*dmhalo.m_triangle*(dmhalo.r_triangle**2)/tracer.ltotnorm/tracer.r_scale**3
         mu=g*(dmhalo.enclosed_mass(params['lamb']*tracer.rhalf_2d/dmhalo.r_triangle)+tracer.enclosed_luminosity(params['lamb']*tracer.rhalf_2d/tracer.r_scale)*tracer.luminosity_tot*tracer.upsilon/dmhalo.m_triangle)*dmhalo.m_triangle/params['lamb']/tracer.rhalf_2d/vvar
-    return np.sqrt(vvar),mu.value
+        return np.sqrt(vvar),mu
